unit TimePanel;

{
  Copyright Notice
  ------------------------------------------------------------------------------
  TTimePanel is  Copyright 1999 Dave Hughes
  EMail : dave@waveform.free-online.co.uk
          hughesd6@cs.man.ac.uk
          dave@hartlib.demon.co.uk
  Web :   http://www.waveform.free-online.co.uk

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  + The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software in a
    product, an acknowledgment in the product documentation would be
    appreciated but is not required.
  + Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.
  + This notice may not be removed or altered from any source distribution.
}

{
  Alteration History (* Information, + Added Functionality, - Bug Fixed)
  ------------------------------------------------------------------------------

  08/10/99 v1.0a
  * First release into the public domain!

  15/10/99 v1.1
  + Thanks ever so much for the alterations to the scrolling routines by
    David Johnson of Futonics (marked DJ in this file), control boxes hitting
    the edge of the control now cause scrolling to occur

  18/10/99 v1.2
  * Tidied up some typos following a stupid botch up with a global
    Find/Replace

  21/10/99 v1.3
  - Corrected problem of hiding the start and finish boxes when duration is
    small by shifting duration box to middle of the "stack" of boxes

  22/10/99 v2.0
  + Added Pause time capability
  + Added Align property
  + Added standard TPanel events
  + Added Simple time capability (one time only - no duration / pause etc.),
    thanks again to David Johnson for the idea (and a hint for implementation)!
  * Removed TimeFormat, and replaced with extra properties StartFormat and
    FinishFormat
  * Restructured property default code to allow for easier alteration
  * Optimised the drag code
  * Removed unnecessarily declared type
  * Restructured class hierarchy with TCustomTimePanel
  - Fixed problems with setting StartTime and FinishTime when parameter has
    date information included
  - Removed useless TabStop and TabOrder properties (component cannot receive
    keyboard focus, so these are pointless)
  - Removed useless Color property (cannot currently change background colour)
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

{ Component defaults }
const
  TP_DefBevelInner    = bvNone;
  TP_DefBevelOuter    = bvRaised;
  TP_DefBevelWidth    = 1;
  TP_DefBorderStyle   = bsNone;
  TP_DefBorderWidth   = 0;
  TP_DefColor         = clBtnFace;
  TP_DefDurationColor = clYellow;
  TP_DefDurationFormat= 'D %d:%.2d';
  TP_DefFinishColor   = clRed;
  TP_DefFinishDay     = 1;
  TP_DefFinishFormat  = '"F" HH:mm';
  TP_DefFinishHours   = 11;
  TP_DefFinishMins    = 30;
  TP_DefPauseColor    = clAqua;
  TP_DefPauseFormat   = 'P %d:%.2d';
  TP_DefPauseHours    = 0;
  TP_DefPauseMins     = 30;
  TP_DefSelColor      = clTeal;
  TP_DefShowConnect   = True;
  TP_DefShowDays      = 1;
  TP_DefShowDuration  = True;
  TP_DefShowPause     = True;
  TP_DefSimpleColor   = clLime;
  TP_DefSimpleFormat  = 'HH:mm';
  TP_DefSimpleHours   = 9;
  TP_DefSimpleMins    = 0;
  TP_DefSimplePanel   = False;
  TP_DefStartColor    = clLime;
  TP_DefStartFormat   = '"S" HH:mm';
  TP_DefStartHours    = 9;
  TP_DefStartMins     = 0;
  TP_DefTimeColor     = clGray;
  TP_DefTopDay        = 1;
  TP_DefTopHour       = 7;
  TP_DefTopMin        = 0;
  TP_DefWorkFinish    = 17;
  TP_DefWorkStart     = 9;
  TP_DefWorkTimeColor = clBlack;
  TP_DefWidth         = 128;
  TP_DefHeight        = 200;

type

  { Exception for TTimeBoxControl Create method }

  ENotATimePanel = class(Exception); // Shouldn't be needed (ever!)

  { Custom types used by TTimePanel }

  TTPDays = 1..2;
  TTPHours = 0..23;
  TTPDoubleHours = 0..47;
  TTPMins = 0..59;
  TTPArrowDirection = (adUp, adDown);
  TTPScrollDirection = (sdUp, sdDown, sdNotScrolling);
  TTPControlBox = (cbStart, cbFinish, cbPause, cbDuration, cbSimple);

  { TTimeBoxControl }

  TTimeBoxControl = class(TGraphicControl)
  protected
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    property OnMouseMove;
  end;

  { TCustomTimePanel }

  TCustomTimePanel = class(TCustomPanel)
  private
    { Private declarations }
    FBottomArrowRect: TRect;
    FBottomDay: TTPDays;
    FBottomHour: TTPHours;
    FDurationColor: TColor;
    FDurationControl: TTimeBoxControl;
    FDurationFormat: String;
    FDurationRect: TRect;
    FDurationVisible: Boolean;
    FFinishColor: TColor;
    FFinishControl: TTimeBoxControl;
    FFinishDay: TTPDays;
    FFinishFormat: String;
    FFinishHours: TTPHours;
    FFinishMins: TTPMins;
    FFinishRect: TRect;
    FFinishVisible: Boolean;
    FPauseColor: TColor;
    FPauseControl: TTimeBoxControl;
    FPauseFormat: String;
    FPauseHours: TTPHours;
    FPauseMins: TTPMins;
    FPauseRect: TRect;
    FPauseVisible: Boolean;
    FTextHeight: Integer;
    FTimeColor: TColor;
    FTopArrowRect: TRect;
    FTopDay: TTPDays;
    FTopHour: TTPHours;
    FTopMin: TTPMins;
    FScrollAmount: Single;
    FScrolling: TTPScrollDirection;
    FSelColor: TColor;
    FShowConnect: Boolean;
    FShowDays: TTPDays;
    FShowDuration: Boolean;
    FShowPause: Boolean;
    FSimpleColor: TColor;
    FSimpleControl: TTimeBoxControl;
    FSimpleFormat: String;
    FSimpleHours: TTPHours;
    FSimpleMins: TTPMins;
    FSimplePanel: Boolean;
    FSimpleRect: TRect;
    FSimpleVisible: Boolean;
    FStartColor: TColor;
    FStartControl: TTimeBoxControl;
    FStartFormat: String;
    FStartHours: TTPHours;
    FStartMins: TTPMins;
    FStartRect: TRect;
    FStartVisible: Boolean;
    FWorkFinish: TTPHours;
    FWorkStart: TTPHours;
    FWorkTimeColor: TColor;
    FOnTimeChange: TNotifyEvent;
    function GetDurationHours: TTPHours;
    function GetDurationMins: TTPMins;
    function GetFinishTime: TDateTime;
    function GetStartTime: TDateTime;
    function GetSimpleTime: TDateTime;
    function IsInRect(ARect: TRect; X, Y: Integer): Boolean;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer);
      override;
    procedure StartMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure FinishMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure DurationMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure PauseMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure SimpleMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;
    procedure SetDurationColor(Value: TColor);
    procedure SetDurationFormat(Value: String);
    procedure SetDurationHours(Value: TTPHours);
    procedure SetDurationMins(Value: TTPMins);
    procedure SetFinishColor(Value: TColor);
    procedure SetFinishDay(Value: TTPDays);
    procedure SetFinishFormat(Value: String);
    procedure SetFinishHours(Value: TTPHours);
    procedure SetFinishMins(Value: TTPMins);
    procedure SetFinishTime(Value: TDateTime);
    procedure SetPauseColor(Value: TColor);
    procedure SetPauseFormat(Value: String);
    procedure SetPauseHours(Value: TTPHours);
    procedure SetPauseMins(Value: TTPMins);
    procedure SetSelColor(Value: TColor);
    procedure SetSimpleColor(Value: TColor);
    procedure SetSimpleFormat(Value: String);
    procedure SetSimpleHours(Value: TTPHours);
    procedure SetSimpleMins(Value: TTPMins);
    procedure SetSimplePanel(Value: Boolean);
    procedure SetSimpleTime(Value: TDateTime);
    procedure SetShowConnect(Value: Boolean);
    procedure SetShowDays(Value: TTPDays);
    procedure SetShowDuration(Value: Boolean);
    procedure SetShowPause(Value: Boolean);
    procedure SetStartColor(Value: TColor);
    procedure SetStartFormat(Value: String);
    procedure SetStartHours(Value: TTPHours);
    procedure SetStartMins(Value: TTPMins);
    procedure SetStartTime(Value: TDateTime);
    procedure SetTimeColor(Value: TColor);
    procedure SetTopDay(Value: TTPDays);
    procedure SetTopHour(Value: TTPHours);
    procedure SetTopMin(Value: TTPMins);
    procedure SetWorkFinish(Value: TTPHours);
    procedure SetWorkStart(Value: TTPHours);
    procedure SetWorkTimeColor(Value: TColor);
    procedure Scroll(Dir: TTPScrollDirection; Hours: TTPHours; Mins: TTPMins);
    procedure DoTimeChange;
  protected
    { Protected declarations }
    procedure CalcRange;
    procedure DrawArrow(C: TCanvas; Dir: TTPArrowDirection; AColor: TColor);
    procedure DrawConnect(C: TCanvas);
    procedure DrawFrame(C: TCanvas);
    function DrawTimeBox(C: TCanvas; Day: TTPDays; Hours: TTPHours;
      Mins: TTPMins; AColor: TColor; BoxText: String): TRect;
    procedure DrawTimeBoxes(C: TCanvas);
    procedure DrawTimeNumbers(C: TCanvas);
    procedure Paint; override;
    property DurationColor: TColor read FDurationColor write SetDurationColor default TP_DefDurationColor;
    property DurationFormat: String read FDurationFormat write SetDurationFormat;
    property DurationHours: TTPHours read GetDurationHours write SetDurationHours stored False;
    property DurationMins: TTPMins read GetDurationMins write SetDurationMins stored False;
    property DurationVisible: Boolean read FDurationVisible;
    property FinishColor: TColor read FFinishColor write SetFinishColor default TP_DefFinishColor;
    property FinishDay: TTPDays read FFinishDay write SetFinishDay default TP_DefFinishDay;
    property FinishFormat: String read FFinishFormat write SetFinishFormat;
    property FinishHours: TTPHours read FFinishHours write SetFinishHours default TP_DefFinishHours;
    property FinishMins: TTPMins read FFinishMins write SetFinishMins default TP_DefFinishMins;
    property FinishVisible: Boolean read FFinishVisible;
    property PauseColor: TColor read FPauseColor write SetPauseColor default TP_DefPauseColor;
    property PauseFormat: String read FPauseFormat write SetPauseFormat;
    property PauseHours: TTPHours read FPauseHours write SetPauseHours default TP_DefPauseHours;
    property PauseMins: TTPMins read FPauseMins write SetPauseMins default TP_DefPauseMins;
    property PauseVisible: Boolean read FPauseVisible;
    property SelColor: TColor read FSelColor write SetSelColor default TP_DefSelColor;
    property ShowConnect: Boolean read FShowConnect write SetShowConnect default TP_DefShowConnect;
    property ShowDays: TTPDays read FShowDays write SetShowDays default TP_DefShowDays;
    property ShowDuration: Boolean read FShowDuration write SetShowDuration default TP_DefShowDuration;
    property ShowPause: Boolean read FShowPause write SetShowPause default TP_DefShowPause;
    property SimpleColor: TColor read FSimpleColor write SetSimpleColor default TP_DefSimpleColor;
    property SimpleFormat: String read FSimpleFormat write SetSimpleFormat;
    property SimpleHours: TTPHours read FSimpleHours write SetSimpleHours default TP_DefSimpleHours;
    property SimpleMins: TTPMins read FSimpleMins write SetSimpleMins default TP_DefSimpleMins;
    property SimplePanel: Boolean read FSimplePanel write SetSimplePanel default TP_DefSimplePanel;
    property SimpleVisible: Boolean read FSimpleVisible;
    property StartColor: TColor read FStartColor write SetStartColor default TP_DefStartColor;
    property StartFormat: String read FStartFormat write SetStartFormat;
    property StartHours: TTPHours read FStartHours write SetStartHours default TP_DefStartHours;
    property StartMins: TTPMins read FStartMins write SetStartMins default TP_DefStartMins;
    property StartVisible: Boolean read FStartVisible;
    property TimeColor: TColor read FTimeColor write SetTimeColor default TP_DefTimeColor;
    property TopDay: TTPDays read FTopDay write SetTopDay default TP_DefTopDay;
    property TopHour: TTPHours read FTopHour write SetTopHour default TP_DefTopHour;
    property TopMin: TTPMins read FTopMin write SetTopMin default TP_DefTopMin;
    property Width default TP_DefWidth;
    property WorkFinish: TTPHours read FWorkFinish write SetWorkFinish default TP_DefWorkFinish;
    property WorkStart: TTPHours read FWorkStart write SetWorkStart default TP_DefWorkStart;
    property WorkTimeColor: TColor read FWorkTimeColor write SetWorkTimeColor default TP_DefWorkTimeColor;
    property OnTimeChange: TNotifyEvent read FOnTimeChange write FOnTimeChange;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure IncTime(ControlBox: TTPControlBox; Hours: TTPHours;
      Mins: TTPMins);
    procedure DecTime(ControlBox: TTPControlBox; Hours: TTPHours;
      Mins: TTPMins);
    property FinishTime: TDateTime read GetFinishTime write SetFinishTime;
    property SimpleTime: TDateTime read GetSimpleTime write SetSimpleTime;
    property StartTime: TDateTime read GetStartTime write SetStartTime;
  end;

  TTimePanel = class(TCustomTimePanel)
  published
    { Published declarations }
    property Align;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderStyle;
    property BorderWidth;
    property Cursor;
    property DurationColor;
    property DurationFormat;
    property DurationHours;
    property DurationMins;
    property DurationVisible;
    property FinishColor;
    property FinishDay;
    property FinishFormat;
    property FinishHours;
    property FinishMins;
    property FinishVisible;
    property Font;
    property Height;
    property ParentShowHint;
    property PauseColor;
    property PauseFormat;
    property PauseHours;
    property PauseMins;
    property PauseVisible;
    property SelColor;
    property ShowConnect;
    property ShowDays;
    property ShowDuration;
    property ShowPause;
    property ShowHint;
    property SimpleColor;
    property SimpleFormat;
    property SimpleHours;
    property SimpleMins;
    property SimplePanel;
    property SimpleVisible;
    property StartColor;
    property StartFormat;
    property StartHours;
    property StartMins;
    property StartVisible;
    property TimeColor;
    property TopDay;
    property TopHour;
    property TopMin;
    property Width;
    property WorkFinish;
    property WorkStart;
    property WorkTimeColor;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
    property OnTimeChange;
  end;

procedure Register;

implementation

{ TTimeBoxControl }

constructor TTimeBoxControl.Create(AOwner: TComponent);
begin
  if AOwner is TTimePanel then
  begin
    inherited Create(AOwner);
    ControlStyle := ControlStyle - [csOpaque];
    Visible := True;
  end
  else
    raise ENotATimePanel.Create('Parent is not a TTimePanel');
end;

{ TTimePanel }

constructor TCustomTimePanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];

  // Set defaults
  FTopDay := TP_DefTopDay;
  FTopHour := TP_DefTopHour;
  FTopMin := TP_DefTopMin;
  FShowConnect := TP_DefShowConnect;
  FShowDays := TP_DefShowDays;
  FSimplePanel := TP_DefSimplePanel;
  FSimpleHours := TP_DefSimpleHours;
  FSimpleMins := TP_DefSimpleMins;
  FSimpleColor := TP_DefSimpleColor;
  FSimpleControl := TTimeBoxControl.Create(Self);
  FSimpleControl.Parent := Self;
  FSimpleControl.OnMouseMove := SimpleMouseMove;
  FSimpleFormat := TP_DefSimpleFormat;
  FShowDuration := TP_DefShowDuration;
  FDurationColor := TP_DefDurationColor;
  FDurationControl := TTimeBoxControl.Create(Self);
  FDurationControl.Parent := Self;
  FDurationControl.OnMouseMove := DurationMouseMove;
  FDurationFormat := TP_DefDurationFormat;
  FShowPause := TP_DefShowPause;
  FPauseHours := TP_DefPauseHours;
  FPauseMins := TP_DefPauseMins;
  FPauseColor := TP_DefPauseColor;
  FPauseControl := TTimeBoxControl.Create(Self);
  FPauseControl.Parent := Self;
  FPauseControl.OnMouseMove := PauseMouseMove;
  FPauseFormat := TP_DefPauseFormat;
  FStartHours := TP_DefStartHours;
  FStartMins := TP_DefStartMins;
  FStartColor := TP_DefStartColor;
  FStartFormat := TP_DefStartFormat;
  FStartControl := TTimeBoxControl.Create(Self);
  FStartControl.Parent := Self;
  FStartControl.OnMouseMove := StartMouseMove;
  FFinishDay := TP_DefFinishDay;
  FFinishHours := TP_DefFinishHours;
  FFinishMins := TP_DefFinishMins;
  FFinishColor := TP_DefFinishColor;
  FFinishFormat := TP_DefFinishFormat;
  FFinishControl := TTimeBoxControl.Create(Self);
  FFinishControl.Parent := Self;
  FFinishControl.OnMouseMove := FinishMouseMove;
  FTimeColor := TP_DefTimeColor;
  FWorkTimeColor := TP_DefWorkTimeColor;
  FWorkStart := TP_DefWorkStart;
  FWorkFinish := TP_DefWorkFinish;
  FSelColor := TP_DefSelColor;
  FScrolling := sdNotScrolling;

  // Set inherited properties defaults
  Caption := '';
  Width := TP_DefWidth;
  Height := TP_DefHeight;
  BevelInner := TP_DefBevelInner;
  BevelOuter := TP_DefBevelOuter;
  BevelWidth := TP_DefBevelWidth;
  BorderStyle := TP_DefBorderStyle;
  BorderWidth := TP_DefBorderWidth;
  Color := TP_DefColor;
  TabStop := False;
  Font.Name := 'MS Sans Serif';
  Font.Size := 8;
  Font.Style := [];
end;

{------------------------------------------------------------ Paint Handlers --}

const
  ArrowHeight = 12;
  BoxLeftMargin = 18;
  SingleBorderMargin = 5;
  TimeNumMargin = 4;

procedure TCustomTimePanel.DrawArrow(C: TCanvas; Dir: TTPArrowDirection;
  AColor: TColor);
var
  Arrow: Array [1..3] of TPoint;
  Margin: Integer;
begin
  with C do
  begin
    Brush.Color := AColor;
    Brush.Style := bsSolid;
    Pen.Color := AColor;
    Pen.Style := psSolid;
  end;
  if Dir = adUp then
  begin
    Arrow[1] := Point(2, 8);
    Arrow[2] := Point(8, 2);
    Arrow[3] := Point(14, 8);
    with FTopArrowRect do
    begin
      Top := 2;
      Left := 2;
      Right := 14;
      Bottom := 8;
    end;
  end
  else
  begin
    if BorderStyle = bsSingle then
      Margin := SingleBorderMargin
    else
     Margin := 0;
    Arrow[1] := Point(2, Height - (8 + Margin));
    Arrow[2] := Point(8, Height - (2 + Margin));
    Arrow[3] := Point(14, Height - (8 + Margin));
    with FBottomArrowRect do
    begin
      Top := Height - (8 + Margin);
      Left := 2;
      Bottom := Height - (2 + Margin);
      Right := 14;
    end;
  end;
  C.Polygon(Arrow);
end;

procedure TCustomTimePanel.DrawConnect(C: TCanvas);
const
  LineWidth = 4;
var
  Hours, TopHours: TTPDoubleHours;
  Mins: TTPMins;
  LineTop, LineBottom, LineLeft: Integer;
begin
  CalcRange;
  TopHours := (FTopHour + ((FTopDay - 1) * 24));
  with C do
  begin
    // Set up pen
    Pen.Style := psSolid;
    Pen.Width := LineWidth;
    LineLeft := Trunc(TextWidth('88:88')) + BoxLeftMargin + LineWidth div 2;
    // Draw from Start to Pause
    Hours := FStartHours;
    Mins := FStartMins;
    LineTop := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
     (Trunc(FTextHeight * (FTopMin / 60.0))) +
     (Trunc(FTextHeight * (Mins / 60.0))) +
     (TextHeight('88:88') div 2);
    Hours := FStartHours + FPauseHours;
    Mins := FStartMins + FPauseMins;
    LineBottom := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
     (Trunc(FTextHeight * (FTopMin / 60.0))) +
     (Trunc(FTextHeight * (Mins / 60.0))) +
     (TextHeight('88:88') div 2);
    Pen.Color := clGray;
    MoveTo(LineLeft, LineTop);
    LineTo(LineLeft, LineBottom);
    // Draw from Pause to Finish
    Hours := FStartHours + FPauseHours;
    Mins := FStartMins + FPauseMins;
    LineTop := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
     (Trunc(FTextHeight * (FTopMin / 60.0))) +
     (Trunc(FTextHeight * (Mins / 60.0))) +
     (TextHeight('88:88') div 2);
    Hours := FFinishHours + ((FFinishDay - 1) * 24);
    Mins := FFinishMins;
    LineBottom := ArrowHeight + ((Hours - TopHours) * FTextHeight) -
     (Trunc(FTextHeight * (FTopMin / 60.0))) +
     (Trunc(FTextHeight * (Mins / 60.0))) +
     (TextHeight('88:88') div 2);
    Pen.Color := clBlack;
    MoveTo(LineLeft, LineTop);
    LineTo(LineLeft, LineBottom);
    // Reset pen
    Pen.Width := 1;
  end;
end;

function TCustomTimePanel.DrawTimeBox(C: TCanvas; Day: TTPDays; Hours: TTPHours;
  Mins: TTPMins; AColor: TColor; BoxText: String): TRect;
const
  VMargin = 2;
  HMargin = 8;
var
  BoxRect: TRect;
  RHours, RTopHours: TTPDoubleHours;
begin
  CalcRange;
  RTopHours := (FTopHour + ((FTopDay - 1) * 24));
  RHours := (Hours + ((Day - 1) * 24));
  with C do
  begin
    Brush.Color := AColor;
    Brush.Style := bsSolid;
    Pen.Color := clBlack;
    Pen.Style := psSolid;
    BoxRect.Left := TextWidth('88:88') + BoxLeftMargin;
    BoxRect.Right := BoxRect.Left + TextWidth(BoxText) + (2 * HMargin);
    BoxRect.Top := 12 + ((RHours - RTopHours) * FTextHeight) -
      (Trunc(FTextHeight * (FTopMin / 60.0))) + (Trunc(FTextHeight *
      (Mins / 60.0))) - VMargin;
    BoxRect.Bottom := BoxRect.Top + TextHeight(BoxText) + VMargin;
    with BoxRect do
      Rectangle(Left, Top, Right, Bottom);
    Brush.Style := bsClear;
    Font.Color := clBlack;
    TextOut(BoxRect.Left + HMargin, BoxRect.Top + (VMargin div 2), BoxText);
  end;
  Result := BoxRect;
end;

procedure TCustomTimePanel.DrawTimeBoxes(C: TCanvas);
var
  StMins, FinMins, PMins, DMins: Integer;
  DurDay, PaDay: TTPDays;
  DurHours, PaHours: TTPHours;
  DurMins, PaMins: TTPMins;
begin
  if FSimplePanel then
  begin
    // Simple time box
    FSimpleRect := DrawTimeBox(C, 1, FSimpleHours, FSimpleMins, FSimpleColor,
      FormatDateTime(FSimpleFormat, SimpleTime));
    with FSimpleRect do
    begin
      if (Bottom > 0) and (Top < Height) then
        FSimpleVisible := True
      else
        FSimpleVisible := False;
      FSimpleControl.Top := Top;
      FSimpleControl.Left := Left;
      FSimpleControl.Height := (Bottom - Top);
      FSimpleControl.Width := (Right - Left);
    end;
  end
  else
  begin
    // Duration time box
    if FShowDuration then
    begin
      StMins := ((FStartHours * 60) + FStartMins);
      FinMins := (((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
        FFinishMins);
      PMins := StMins + ((FPauseHours * 60) + FPauseMins);
      DMins := PMins + ((FinMins - PMins) div 2);
      DurDay := (DMins div (24 * 60)) + 1;
      DurHours := (DMins div 60) mod 24;
      DurMins := DMins mod 60;
      FDurationRect := DrawTimeBox(C, DurDay, DurHours, DurMins,
        FDurationColor, Format(FDurationFormat, [GetDurationHours,
        GetDurationMins]));
      with FDurationRect do
      begin
        if (Bottom > 0) and (Top < Height) then
          FDurationVisible := True
        else
          FDurationVisible := False;
        FDurationControl.Top := Top;
        FDurationControl.Left := Left;
        FDurationControl.Height := (Bottom - Top);
        FDurationControl.Width := (Right - Left);
      end;
    end
    else
      FDurationVisible := False;
    // Pause time box
    if FShowPause then
    begin
      StMins := ((FStartHours * 60) + FStartMins);
      PMins := StMins + ((FPauseHours * 60) + FPauseMins);
      PaDay := (PMins div (24 * 60)) + 1;
      PaHours := (PMins div 60) mod 24;
      PaMins := PMins mod 60;
      FPauseRect := DrawTimeBox(C, PaDay, PaHours, PaMins, FPauseColor,
        Format(FPauseFormat, [FPauseHours, FPauseMins]));
      with FPauseRect do
      begin
        if (Bottom > 0) and (Top < Height) then
          FPauseVisible := True
        else
          FPauseVisible := False;
        FPauseControl.Top := Top;
        FPauseControl.Left := Left;
        FPauseControl.Height := (Bottom - Top);
        FPauseControl.Width := (Right - Left);
      end;
    end
    else
      FPauseVisible := False;
    // Start time box
    FStartRect := DrawTimeBox(C, 1, FStartHours, FStartMins, FStartColor,
      FormatDateTime(FStartFormat, StartTime));
    with FStartRect do
    begin
      if (Bottom > 0) and (Top < Height) then
        FStartVisible := True
      else
        FStartVisible := False;
      FStartControl.Top := Top;
      FStartControl.Left := Left;
      FStartControl.Height := (Bottom - Top);
      FStartControl.Width := (Right - Left);
    end;
    // Finish time box
    FFinishRect := DrawTimeBox(C, FFinishDay, FFinishHours, FFinishMins,
      FFinishColor, FormatDateTime(FFinishFormat, FinishTime));
    with FFinishRect do
    begin
      if (Bottom > 0) and (Top < Height) then
        FFinishVisible := True
      else
        FFinishVisible := False;
      FFinishControl.Top := Top;
      FFinishControl.Left := Left;
      FFinishControl.Height := (Bottom - Top);
      FFinishControl.Width := (Right - Left);
    end;
  end;
end;

procedure TCustomTimePanel.DrawTimeNumbers(C: TCanvas);
var
  LineRect, TextClipRect: TRect;
  TextTop: Integer;
  HourNum, RTopHours, RBottomHours: TTPDoubleHours;
begin
  C.Font := Font;
  with C do
  begin
    Brush.Style := bsClear;
    with TextClipRect do
    begin
      Left := 0;
      Top := ArrowHeight;
      Right := TextWidth('88:88') + (2 * TimeNumMargin);
      if BorderStyle = bsSingle then
        Bottom := Height - ArrowHeight - SingleBorderMargin
      else
        Bottom := Height - ArrowHeight;
    end;
    CalcRange;
    RTopHours := (FTopHour + ((FTopDay - 1) * 24));
    RBottomHours := (FBottomHour + ((FBottomDay - 1) * 24));
    for HourNum := RTopHours to RBottomHours do
    begin
      TextTop := TextClipRect.Top + ((HourNum - RTopHours) * FTextHeight) -
        (Trunc(FTextHeight * (FTopMin / 60.0)));
      if ((HourNum mod 24) >= FWorkStart) and
        ((HourNum mod 24) <= FWorkFinish) then
        Font.Color := FWorkTimeColor
      else
        Font.Color := FTimeColor;
      TextRect(TextClipRect, TimeNumMargin, TextTop, Format('%.2d:00',
        [HourNum mod 24]));
    end;
  end;
  LineRect := GetClientRect;
  LineRect.Left := C.TextWidth('88:88') + (2 * TimeNumMargin);
  LineRect.Right := LineRect.Left + 2;
  Frame3D(C, LineRect, clBtnShadow, clBtnHighlight, 1);
end;

procedure TCustomTimePanel.CalcRange;
var
  Metric: TTextMetric;
  Margin, CalcBottom: Integer;
begin
  Margin := ArrowHeight * 2;
  if BorderStyle = bsSingle then
    Margin := Margin + SingleBorderMargin;
  GetTextMetrics(Canvas.Handle, Metric);
  FTextHeight := Metric.tmHeight * 2;
  CalcBottom := (FTopHour + ((FTopDay - 1) * 24)) +
    ((Height - Margin) div FTextHeight) + 1;
  if CalcBottom > 47 then
  begin
    FBottomDay := FShowDays;
    FBottomHour := 23;
  end
  else
  begin
    if CalcBottom > 23 then
    begin
      FBottomDay := FShowDays;
      case FShowDays of
        1: FBottomHour := 23;
        2: FBottomHour := CalcBottom;
      end;
    end
    else
    begin
      FBottomDay := 1;
      FBottomHour := CalcBottom;
    end;
  end;
end;

procedure TCustomTimePanel.DrawFrame(C: TCanvas);
var
  Rect: TRect;
  TopColor, BottomColor: TColor;

  procedure AdjustColors(Bevel: TPanelBevel);
  begin
    TopColor := clBtnHighlight;
    if Bevel = bvLowered then TopColor := clBtnShadow;
    BottomColor := clBtnShadow;
    if Bevel = bvLowered then BottomColor := clBtnHighlight;
  end;

begin
  Rect := GetClientRect;
  if BevelOuter <> bvNone then
  begin
    AdjustColors(BevelOuter);
    Frame3D(C, Rect, TopColor, BottomColor, BevelWidth);
  end;
  Frame3D(C, Rect, Color, Color, BorderWidth);
  if BevelInner <> bvNone then
  begin
    AdjustColors(BevelInner);
    Frame3D(C, Rect, TopColor, BottomColor, BevelWidth);
  end;
  with C do
  begin
    Brush.Color := Color;
    FillRect(Rect);
  end;
end;

procedure TCustomTimePanel.Paint;
var
  B: TBitmap;
begin
  B := TBitmap.Create;
  try
    B.Width := Width;
    B.Height := Height;
    Canvas.Font := Font;
    B.Canvas.Font := Font;
    // Draw frame
    DrawFrame(B.Canvas);
    // Draw left hand side numbers
    DrawTimeNumbers(B.Canvas);
    // Draw connecting line
    if (not FSimplePanel) and FShowConnect then
      DrawConnect(B.Canvas);
    // Draw (start, finish, duration and pause) or (simple) boxes
    DrawTimeBoxes(B.Canvas);
    // Draw arrows
    if FScrolling = sdNotScrolling then
    begin
      DrawArrow(B.Canvas, adUp, clBlack);
      DrawArrow(B.Canvas, adDown, clBlack);
    end
    else
    begin
      if FScrolling = sdUp then
      begin
        DrawArrow(B.Canvas, adUp, FSelColor);
        DrawArrow(B.Canvas, adDown, clBlack);
      end
      else
      begin
        DrawArrow(B.Canvas, adUp, clBlack);
        DrawArrow(B.Canvas, adDown, FSelColor);
      end;
    end;
    // Draw temporary bitmap onto control canvas
    Canvas.Draw(0, 0, B);
  finally
    // Destroy temporary bitmap
    B.Free;
  end;
  // Deal with scrolling
  if FScrolling <> sdNotScrolling then
  begin
    if FScrolling = sdUp then
    begin
      Scroll(sdUp, 0, Trunc(FScrollAmount));
      Invalidate;
      Sleep(5);
    end
    else
    begin
      Scroll(sdDown, 0, Trunc(FScrollAmount));
      Invalidate;
      Sleep(5);
    end;
    FScrollAmount := FScrollAmount + 0.5;
    if FScrollAmount > 10.0 then // 15/10/99 DJ
      FScrollAmount := 10.0;  // 15/10/99 DJ
  end;
end;

{--------------------------------------------------------- Colour Properties --}

procedure TCustomTimePanel.SetDurationColor(Value: TColor);
begin
  if (not FSimplePanel) and (Value <> FDurationColor) then
  begin
    FDurationColor := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetStartColor(Value: TColor);
begin
  if (not FSimplePanel) and (Value <> FStartColor) then
  begin
    FStartColor := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetFinishColor(Value: TColor);
begin
  if (not FSimplePanel) and (Value <> FFinishColor) then
  begin
    FFinishColor := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetPauseColor(Value: TColor);
begin
  if (not FSimplePanel) and (Value <> FPauseColor) then
  begin
    FPauseColor := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetTimeColor(Value: TColor);
begin
  if Value <> FTimeColor then
  begin
    FTimeColor := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetWorkTimeColor(Value: TColor);
begin
  if Value <> FWorkTimeColor then
  begin
    FWorkTimeColor := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetSelColor(Value: TColor);
begin
  if Value <> FSelColor then
  begin
    FSelColor := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetSimpleColor(Value: TColor);
begin
  if (FSimplePanel) and (Value <> FSimpleColor) then
  begin
    FSimpleColor := Value;
    Invalidate;
  end;
end;

{----------------------------------------------------------- Time Properties --}

procedure TCustomTimePanel.SetDurationHours(Value: TTPHours);
var
  FinMins: Integer;
  CalcFinishMins: TTPMins;
  CalcFinishHours: TTPHours;
  CalcFinishDay: TTPDays;
begin
  if (not FSimplePanel) and (Value <> GetDurationHours) then
  begin
    FinMins := (FStartHours * 60) + FStartMins +
      (FPauseHours * 60) + FPauseMins + (Value * 60) + GetDurationMins;
    CalcFinishMins := FinMins mod 60;
    CalcFinishHours := (FinMins div 60) mod 24;
    CalcFinishDay := (FinMins div (24 * 60)) + 1;
    if Value < GetDurationHours then
    begin
      FFinishDay := CalcFinishDay;
      FFinishHours := CalcFinishHours;
      FFinishMins := CalcFinishMins;
    end
    else
    begin
      if CalcFinishDay > FShowDays then
      begin
        FFinishDay := 1;
        FFinishHours := 23;
        FFinishMins := 59;
      end
      else
      begin
        FFinishDay := CalcFinishDay;
        FFinishHours := CalcFinishHours;
        FFinishMins := CalcFinishMins;
      end;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetDurationMins(Value: TTPMins);
var
  FinMins: Integer;
  CalcFinishMins: TTPMins;
  CalcFinishHours: TTPHours;
  CalcFinishDay: TTPDays;
begin
  if (not FSimplePanel) and (Value <> GetDurationMins) then
  begin
    FinMins := (FStartHours * 60) + FStartMins + (FPauseHours * 60) +
      FPauseMins + (GetDurationHours * 60) + Value;
    CalcFinishMins := FinMins mod 60;
    CalcFinishHours := (FinMins div 60) mod 24;
    CalcFinishDay := (FinMins div (24 * 60)) + 1;
    if Value < GetDurationMins then
    begin
      FFinishDay := CalcFinishDay;
      FFinishHours := CalcFinishHours;
      FFinishMins := CalcFinishMins;
    end
    else
    begin
      if CalcFinishDay > FShowDays then
      begin
        FFinishDay := 1;
        FFinishHours := 23;
        FFinishMins := 59;
      end
      else
      begin
        FFinishDay := CalcFinishDay;
        FFinishHours := CalcFinishHours;
        FFinishMins := CalcFinishMins;
      end;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

function TCustomTimePanel.GetDurationHours: TTPHours;
var
  StMins, FinMins, PaMins: Integer;
begin
  StMins := (FStartHours * 60) + FStartMins;
  FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
    FFinishMins;
  PaMins := (FPauseHours * 60) + FPauseMins;
  Result := (FinMins - StMins - PaMins) div 60;
end;

function TCustomTimePanel.GetDurationMins: TTPMins;
var
  StMins, FinMins, PaMins: Integer;
begin
  StMins := (FStartHours * 60) + FStartMins;
  FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
    FFinishMins;
  PaMins := (FPauseHours * 60) + FPauseMins;
  Result := (FinMins - StMins - PaMins) mod 60;
end;

procedure TCustomTimePanel.SetStartHours(Value: TTPHours);
var
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FStartHours) then
  begin
    StMins := (Value * 60) + FStartMins;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
      FFinishMins;
    PaMins := (FPauseHours * 60) + FPauseMins;
    if StMins > FinMins then
    begin
      FStartHours := FFinishHours;
      FStartMins := FFinishMins;
      FPauseHours := 0;
      FPauseMins := 0;
    end
    else if (StMins + PaMins) > FinMins then
    begin
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
      FPauseHours := (FinMins - StMins) div 60;
      FPauseMins := (FinMins - StMins) mod 60;
    end
    else
    begin
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
      if (FinMins - StMins) >= (24 * 60) then
      begin
        FinMins := StMins + ((24 * 60) - 1);
        FFinishDay := 2;
        FFinishHours := (FinMins div 60) mod 24;
        FFinishMins := FinMins mod 60;
      end;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetStartMins(Value: TTPMins);
var
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FStartMins) then
  begin
    StMins := (FStartHours * 60) + Value;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
      FFinishMins;
    PaMins := (FPauseHours * 60) + FPauseMins;
    if StMins > FinMins then
    begin
      FStartHours := FFinishHours;
      FStartMins := FFinishMins;
      FPauseHours := 0;
      FPauseMins := 0;
    end
    else if (StMins + PaMins) > FinMins then
    begin
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
      FPauseHours := (FinMins - StMins) div 60;
      FPauseMins := (FinMins - StMins) mod 60;
    end
    else
    begin
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
      if (FinMins - StMins) >= (24 * 60) then
      begin
        FinMins := StMins + ((24 * 60) - 1);
        FFinishDay := 2;
        FFinishHours := (FinMins div 60) mod 24;
        FFinishMins := FinMins mod 60;
      end;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetFinishDay(Value: TTPDays);
var
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FFinishDay) then
  begin
    if Value <= FShowDays then
    begin
      StMins := (FStartHours * 60) + FStartMins;
      FinMins := ((Value - 1) * (24 * 60)) + (FFinishHours * 60) +
        FFinishMins;
      PaMins := (FPauseHours * 60) + FPauseMins;
      if FinMins < StMins then
      begin
        FFinishDay := 1;
        FFinishHours := FStartHours;
        FFinishMins := FStartMins;
        FPauseHours := 0;
        FPauseMins := 0;
      end
      else if FinMins < (StMins + PaMins) then
      begin
        FFinishDay := (FinMins div (24 * 60)) + 1;
        FFinishHours := (FinMins div 60) mod 24;
        FFinishMins := FinMins mod 60;
        FPauseHours := (FinMins - StMins) div 60;
        FPauseMins := (FinMins - StMins) mod 60;
      end
      else
      begin
        FFinishDay := (FinMins div (24 * 60)) + 1;
        FFinishHours := (FinMins div 60) mod 24;
        FFinishMins := FinMins mod 60;
        if (FinMins - StMins) >= (24 * 60) then
        begin
          StMins := FinMins - ((24 * 60) - 1);
          FStartHours := StMins div 60;
          FStartMins := StMins mod 60;
        end;
      end;
      Invalidate;
      DoTimeChange;
    end;
  end;
end;

procedure TCustomTimePanel.SetFinishHours(Value: TTPHours);
var
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FFinishHours) then
  begin
    StMins := (FStartHours * 60) + FStartMins;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (Value * 60) +
      FFinishMins;
    PaMins := (FPauseHours * 60) + FPauseMins;
    if FinMins < StMins then
    begin
      FFinishDay := 1;
      FFinishHours := FStartHours;
      FFinishMins := FStartMins;
      FPauseHours := 0;
      FPauseMins := 0;
    end
    else if FinMins < (StMins + PaMins) then
    begin
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
      FPauseHours := (FinMins - StMins) div 60;
      FPauseMins := (FinMins - StMins) mod 60;
    end
    else
    begin
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
      if (FinMins - StMins) >= (24 * 60) then
      begin
        StMins := FinMins - ((24 * 60) - 1);
        FStartHours := StMins div 60;
        FStartMins := StMins mod 60;
      end;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetFinishMins(Value: TTPMins);
var
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FFinishMins) then
  begin
    StMins := (FStartHours * 60) + FStartMins;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
      Value;
    PaMins := (FPauseHours * 60) + FPauseMins;
    if FinMins < StMins then
    begin
      FFinishDay := 1;
      FFinishHours := FStartHours;
      FFinishMins := FStartMins;
      FPauseHours := 0;
      FPauseMins := 0;
    end
    else if FinMins < (StMins + PaMins) then
    begin
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
      FPauseHours := (FinMins - StMins) div 60;
      FPauseMins := (FinMins - StMins) mod 60;
    end
    else
    begin
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
      if (FinMins - StMins) >= (24 * 60) then
      begin
        StMins := FinMins - ((24 * 60) - 1);
        FStartHours := StMins div 60;
        FStartMins := StMins mod 60;
      end;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetPauseHours(Value: TTPHours);
var
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FPauseHours) then
  begin
    StMins := (FStartHours * 60) + FStartMins;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
      FFinishMins;
    PaMins := (Value * 60) + FPauseMins;
    if FinMins < (StMins + PaMins) then
      PaMins := FinMins - StMins;
    FPauseHours := PaMins div 60;
    FPauseMins := PaMins mod 60;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetPauseMins(Value: TTPMins);
var
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FPauseMins) then
  begin
    StMins := (FStartHours * 60) + FStartMins;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
      FFinishMins;
    PaMins := (FPauseHours * 60) + Value;
    if FinMins < (StMins + PaMins) then
      PaMins := FinMins - StMins;
    FPauseHours := PaMins div 60;
    FPauseMins := PaMins mod 60;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetSimpleHours(Value: TTPHours);
begin
  if (FSimplePanel) and (Value <> FSimpleHours) then
  begin
    FSimpleHours := Value;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetSimpleMins(Value: TTPMins);
begin
  if (FSimplePanel) and (Value <> FSimpleMins) then
  begin
    FSimpleMins := Value;
    Invalidate;
    DoTimeChange;
  end;
end;

function TCustomTimePanel.GetFinishTime: TDateTime;
begin
  Result := EncodeTime(FFinishHours, FFinishMins, 0, 0);
end;

function TCustomTimePanel.GetStartTime: TDateTime;
begin
  Result := EncodeTime(FStartHours, FStartMins, 0, 0);
end;

function TCustomTimePanel.GetSimpleTime: TDateTime;
begin
  Result := EncodeTime(FSimpleHours, FSimpleMins, 0, 0);
end;

procedure TCustomTimePanel.SetFinishTime(Value: TDateTime);
var
  Hours, Mins, Secs, MSecs: Word;
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Frac(Value) <> GetFinishTime) then
  begin
    DecodeTime(Value, Hours, Mins, Secs, MSecs);
    if Frac(Value) >= GetStartTime then
    begin
      FFinishHours := Hours;
      FFinishMins := Mins;
    end
    else
      if FShowDays = 2 then
      begin
        FFinishDay := 2;
        FFinishHours := Hours;
        FFinishMins := Mins;
      end
      else
      begin
        FFinishHours := FStartHours;
        FFinishMins := FStartMins;
      end;
    StMins := (FStartHours * 60) + FStartMins;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
      FFinishMins;
    PaMins := (FPauseHours * 60) + FPauseMins;
    if FinMins < (StMins + PaMins) then
    begin
      FPauseHours := (FinMins - StMins) div 60;
      FPauseMins := (FinMins - StMins) mod 60;
    end;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetStartTime(Value: TDateTime);
var
  Hours, Mins, Secs, MSecs: Word;
  StMins, FinMins, PaMins: Integer;
begin
  if (not FSimplePanel) and (Frac(Value) <> GetStartTime) then
  begin
    DecodeTime(Value, Hours, Mins, Secs, MSecs);
    if Frac(Value) <= GetFinishTime then
    begin
      FStartHours := Hours;
      FStartMins := Mins;
    end
    else
      if (FShowDays = 2) and (FFinishDay = 2) then
      begin
        FStartHours := Hours;
        FStartMins := Mins;
      end
      else
      begin
        FStartHours := FFinishHours;
        FStartMins := FFinishMins;
      end;
    StMins := (FStartHours * 60) + FStartMins;
    FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) +
      FFinishMins;
    PaMins := (FPauseHours * 60) + FPauseMins;
    if FinMins < (StMins + PaMins) then
    begin
      FPauseHours := (FinMins - StMins) div 60;
      FPauseMins := (FinMins - StMins) mod 60;
    end;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetSimpleTime(Value: TDateTime);
var
  Hours, Mins, Secs, MSecs: Word;
begin
  if (FSimplePanel) and (Frac(Value) <> GetSimpleTime) then
  begin
    DecodeTime(Value, Hours, Mins, Secs, MSecs);
    FSimpleHours := Hours;
    FSimpleMins := Mins;
  end;
end;

{--------------------------------------------------------- Colour Properties --}

procedure TCustomTimePanel.SetTopDay(Value: TTPDays);
begin
  if Value <> FTopDay then
  begin
    if Value <= FShowDays then
    begin
      FTopDay := Value;
      Invalidate;
    end;
  end;
end;

procedure TCustomTimePanel.SetTopHour(Value: TTPHours);
begin
  if Value <> FTopHour then
  begin
    FTopHour := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetTopMin(Value: TTPMins);
begin
  if Value <> FTopMin then
  begin
    FTopMin := Value;
    Invalidate;
  end;
end;

{----------------------------------------------------------- Show Properties --}

procedure TCustomTimePanel.SetShowDays(Value: TTPDays);
var
  StMins, FinMins: Integer;
begin
  if (not FSimplePanel) and (Value <> FShowDays) then
  begin
    FShowDays := Value;
    if (FFinishDay = 2) and (FShowDays = 1) then
    begin
      StMins := ((FStartHours * 60) + FStartMins);
      FinMins := ((FFinishHours * 60) + FFinishMins);
      FFinishDay := 1;
      if FinMins < StMins then
      begin
        FFinishHours := 23;
        FFinishMins := 59;
      end;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetShowDuration(Value: Boolean);
begin
  if (not FSimplePanel) and (Value <> FShowDuration) then
  begin
    FShowDuration := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetShowPause(Value: Boolean);
begin
  if (not FSimplePanel) and (Value <> FShowPause) then
  begin
    FShowPause := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetShowConnect(Value: Boolean);
begin
  if (not FSimplePanel) and (Value <> FShowConnect) then
  begin
    FShowConnect := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetSimplePanel(Value: Boolean);
begin
  if Value <> FSimplePanel then
  begin
    FSimplePanel := Value;
    if FSimplePanel then
      FShowDays := 1;
    Invalidate;
  end;
end;

{--------------------------------------------------------- Format Properties --}

procedure TCustomTimePanel.SetDurationFormat(Value: String);
begin
  if (not FSimplePanel) and (Value <> FDurationFormat) then
  begin
    FDurationFormat := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetPauseFormat(Value: String);
begin
  if (not FSimplePanel) and (Value <> FPauseFormat) then
  begin
    FPauseFormat := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetStartFormat(Value: String);
begin
  if (not FSimplePanel) and (Value <> FStartFormat) then
  begin
    FStartFormat := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetFinishFormat(Value: String);
begin
  if (not FSimplePanel) and (Value <> FStartFormat) then
  begin
    FFinishFormat := Value;
    Invalidate;
  end;
end;

procedure TCustomTimePanel.SetSimpleFormat(Value: String);
begin
  if (FSimplePanel) and (Value <> FSimpleFormat) then
  begin
    FSimpleFormat := Value;
    Invalidate;
  end;
end;

{-------------------------------------------------- Miscellaneous Properties --}

procedure TCustomTimePanel.SetWorkStart(Value: TTPHours);
begin
  if Value <> FWorkStart then
  begin
    if Value <= FWorkFinish then
    begin
      FWorkStart := Value;
    end
    else
    begin
      FWorkStart := FWorkFinish;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

procedure TCustomTimePanel.SetWorkFinish(Value: TTPHours);
begin
  if Value <> FWorkFinish then
  begin
    if Value >= FWorkStart then
    begin
      FWorkFinish := Value;
    end
    else
    begin
      FWorkFinish := FWorkStart;
    end;
    Invalidate;
    DoTimeChange;
  end;
end;

{-------------------------------------------------------- Scrolling Handlers --}

function TCustomTimePanel.IsInRect(ARect: TRect; X, Y: Integer): Boolean;
begin
  with ARect do
    Result := (X >= Left) and (X <= Right) and (Y >= Top) and (Y <= Bottom);
end;

procedure TCustomTimePanel.Scroll(Dir: TTPScrollDirection; Hours: TTPHours;
  Mins: TTPMins);
var
  TopMins, ScrollMins: Integer;
begin
  TopMins := ((FTopDay - 1) * (24 * 60)) + (FTopHour * 60) + FTopMin;
  ScrollMins := (Hours * 60) + Mins;
  if Dir = sdUp then
  begin
    Dec(TopMins, ScrollMins);
    if TopMins < 0 then
      TopMins := 0;
  end
  else
  begin
    Inc(TopMins, ScrollMins);
    if FShowDays = 1 then
    begin
      if TopMins > ((24 * 60) - 1) then
        TopMins := (24 * 60) - 1;
    end
    else
    begin
      if TopMins > ((24 * 60 * 2) - 1) then
        TopMins := (24 * 60 * 2) - 1;
    end;
  end;
  FTopDay := (TopMins div (24 * 60)) + 1;
  FTopHour := (TopMins div 60) mod 24;
  FTopMin := TopMins mod 60;
  Invalidate;
end;

{------------------------------------------------------------- Drag Handlers --}

procedure TCustomTimePanel.IncTime(ControlBox: TTPControlBox; Hours: TTPHours;
  Mins: TTPMins);
var
  SiMins, StMins, FinMins, PaMins, IncMins: Integer;
begin
  SiMins := (FSimpleHours * 60) + FSimpleMins;
  StMins := (FStartHours * 60) + FStartMins;
  FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) + FFinishMins;
  PaMins := (FPauseHours * 60) + FPauseMins;
  IncMins := (Hours * 60) + Mins;
  case ControlBox of
    cbStart:
    begin
      if (StMins + IncMins) > ((24 * 60) - 1) then
        StMins := (24 * 60) - 1
      else
        Inc(StMins, IncMins);
      if StMins > FinMins then
      begin
        StMins := FinMins;
        PaMins := 0;
      end
      else if (StMins + PaMins) > FinMins then
        PaMins := FinMins - StMins;
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
      FPauseHours := PaMins div 60;
      FPauseMins := PaMins mod 60;
    end;
    cbFinish:
    begin
      if (FinMins + IncMins) > ((24 * 60 * FShowDays) - 1) then
        FinMins := (24 * 60 * FShowDays) - 1
      else
        Inc(FinMins, IncMins);
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
    end;
    cbDuration:
    begin
      if (FinMins + IncMins) > ((24 * 60 * FShowDays) - 1) then
        IncMins := ((24 * 60 * FShowDays) - 1) - FinMins;
      if (StMins + IncMins) > ((24 * 60) - 1) then
        IncMins := ((24 * 60) - 1) - StMins;
      Inc(StMins, IncMins);
      Inc(FinMins, IncMins);
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
    end;
    cbPause:
    begin
      if (PaMins + IncMins) > (FinMins - StMins) then
        PaMins := (FinMins - StMins)
      else
        Inc(PaMins, IncMins);
      FPauseHours := PaMins div 60;
      FPauseMins := PaMins mod 60;
    end;
    cbSimple:
    begin
      if (SiMins + IncMins) > ((24 * 60) - 1) then
        SiMins := ((24 * 60) - 1)
      else
        Inc(SiMins, IncMins);
      FSimpleHours := SiMins div 60;
      FSimpleMins := SiMins mod 60;
    end;
  end;
  Invalidate;
  DoTimeChange;
end;

procedure TCustomTimePanel.DecTime(ControlBox: TTPControlBox; Hours: TTPHours;
  Mins: TTPMins);
var
  SiMins, StMins, FinMins, PaMins, DecMins: Integer;
begin
  SiMins := (FSimpleHours * 60) + FSimpleMins;
  StMins := (FStartHours * 60) + FStartMins;
  FinMins := ((FFinishDay - 1) * (24 * 60)) + (FFinishHours * 60) + FFinishMins;
  PaMins := (FPauseHours * 60) + FPauseMins;
  DecMins := (Hours * 60) + Mins;
  case ControlBox of
    cbStart:
    begin
      if (StMins - DecMins) < 0 then
        StMins := 0
      else
        Dec(StMins, DecMins);
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
    end;
    cbFinish:
    begin
      if (FinMins - DecMins) < 0 then
        FinMins := 0
      else
        Dec(FinMins, DecMins);
      if FinMins < StMins then
      begin
        FinMins := StMins;
        PaMins := 0;
      end
      else if (StMins + PaMins) > FinMins then
        PaMins := FinMins - StMins;
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
      FPauseHours := PaMins div 60;
      FPauseMins := PaMins mod 60;
    end;
    cbDuration:
    begin
      if (StMins - DecMins) < 0 then
        DecMins := StMins;
      Dec(StMins, DecMins);
      Dec(FinMins, DecMins);
      FStartHours := StMins div 60;
      FStartMins := StMins mod 60;
      FFinishDay := (FinMins div (24 * 60)) + 1;
      FFinishHours := (FinMins div 60) mod 24;
      FFinishMins := FinMins mod 60;
    end;
    cbPause:
    begin
      if (PaMins - DecMins) < 0 then
        PaMins := 0
      else
        Dec(PaMins, DecMins);
      FPauseHours := PaMins div 60;
      FPauseMins := PaMins mod 60;
    end;
    cbSimple:
    begin
      if (SiMins - DecMins) < 0 then
        SiMins := 0
      else
        Dec(SiMins, DecMins);
      FSimpleHours := SiMins div 60;
      FSimpleMins := SiMins mod 60;
    end;
  end;
  Invalidate;
  DoTimeChange;
end;

procedure TCustomTimePanel.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if IsInRect(FTopArrowRect, X, Y) and (Button = mbLeft) then
  begin
    FScrolling := sdUp;
    FScrollAmount := 1.0;
    Invalidate;
  end
  else if IsInRect(FBottomArrowRect, X, Y) and (Button = mbLeft) then
  begin
    FScrolling := sdDown;
    FScrollAmount := 1.0;
    Invalidate;
  end
  else
    FScrolling := sdNotScrolling;
end;

procedure TCustomTimePanel.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if (FScrolling <> sdNotScrolling) and (Button = mbLeft) then
    FScrolling := sdNotScrolling;
end;

procedure TCustomTimePanel.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  // Deal with scrolling arrows
  if IsInRect(FTopArrowRect, X, Y) then
    DrawArrow(Canvas, adUp, SelColor)
  else
    DrawArrow(Canvas, adUp, clBlack);
  if IsInRect(FBottomArrowRect, X, Y) then
    DrawArrow(Canvas, adDown, SelColor)
  else
    DrawArrow(Canvas, adDown, clBlack);
end;

procedure TCustomTimePanel.StartMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) then
  begin
    OldTop := FStartControl.Top;
    NewTop := FStartControl.Top + Y;
    if Y > FStartControl.Height then
      Dec(NewTop, FStartControl.Height);
    FStartControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * 5;
    if (Y < 0) then
      DecTime(cbStart, Hours, Mins + (FStartMins mod 5))
    else if (Y > FStartControl.Height) then
      IncTime(cbStart, Hours, Mins + (5 - (FStartMins mod 5)));
    // 15/10/99 DJ -------------------------------------------------------------
    if (FStartControl.Top + FStartControl.Height) > Height then
    begin
      FScrolling := sdDown;
      Invalidate;
    end
    else if FStartControl.Top < 0 then
    begin
      FScrolling := sdUp;
      Invalidate;
    end
    else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end;
end;

procedure TCustomTimePanel.FinishMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) then
  begin
    OldTop := FFinishControl.Top;
    NewTop := FFinishControl.Top + Y;
    if Y > FFinishControl.Height then
      Dec(NewTop, FFinishControl.Height);
    FFinishControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * 5;
    if (Y < 0) then
      DecTime(cbFinish, Hours, Mins + (FFinishMins mod 5))
    else if (Y > FFinishControl.Height) then
      IncTime(cbFinish, Hours, Mins + (5 - (FFinishMins mod 5)));
    // 15/10/99 DJ -------------------------------------------------------------
    if (FFinishControl.Top + FFinishControl.Height) > Height then
    begin
      FScrolling := sdDown;
      Invalidate;
    end
    else if FFinishControl.Top < 0 then
    begin
      FScrolling := sdUp;
      Invalidate;
    end else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end;
end;

procedure TCustomTimePanel.DurationMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) and (FShowDuration) then
  begin
    OldTop := FDurationControl.Top;
    NewTop := FDurationControl.Top + Y;
    if Y > FDurationControl.Height then
      Dec(NewTop, FDurationControl.Height);
    FDurationControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * 5;
    if (Y < 0) then
      DecTime(cbDuration, Hours, Mins)
    else if (Y > FDurationControl.Height) then
      IncTime(cbDuration, Hours, Mins);
    // 15/10/99 DJ -------------------------------------------------------------
    if (FDurationControl.Top + FDurationControl.Height) > Height then
    begin
      FScrolling := sdDown;
      Invalidate;
    end
    else if FDurationControl.Top < 0 then
    begin
      FScrolling := sdUp;
      Invalidate;
    end
    else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end;
end;

procedure TCustomTimePanel.PauseMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (not FSimplePanel) and (Shift = [ssLeft]) and (FShowPause) then
  begin
    OldTop := FPauseControl.Top;
    NewTop := FPauseControl.Top + Y;
    if Y > FPauseControl.Height then
      Dec(NewTop, FPauseControl.Height);
    FPauseControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * 5;
    if (Y < 0) then
      DecTime(cbPause, Hours, Mins)
    else if (Y > FPauseControl.Height) then
      IncTime(cbPause, Hours, Mins);
    // 15/10/99 DJ -------------------------------------------------------------
    if (FPauseControl.Top + FPauseControl.Height) > Height then
    begin
      FScrolling := sdDown;
      Invalidate;
    end
    else if FPauseControl.Top < 0 then
    begin
      FScrolling := sdUp;
      Invalidate;
    end
    else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end;
end;

procedure TCustomTimePanel.SimpleMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  OldTop, NewTop: Integer;
  Hours: TTPHours;
  Mins: TTPMins;
begin
  FScrolling := sdNotScrolling; // 15/10/99 DJ
  if (FSimplePanel) and (Shift = [ssLeft]) then
  begin
    OldTop := FSimpleControl.Top;
    NewTop := FSimpleControl.Top + Y;
    if Y > FSimpleControl.Height then
      Dec(NewTop, FSimpleControl.Height);
    FSimpleControl.Top := NewTop;
    Hours := (Abs(OldTop - NewTop) div FTextHeight) mod 24;
    Mins := Trunc((Abs(OldTop - NewTop) / FTextHeight) * 12) * 5;
    if (Y < 0) then
      DecTime(cbSimple, Hours, Mins + (FSimpleMins mod 5))
    else if (Y > FSimpleControl.Height) then
      IncTime(cbSimple, Hours, Mins + (5 - (FSimpleMins mod 5)));
    // 15/10/99 DJ -------------------------------------------------------------
    if (FSimpleControl.Top + FSimpleControl.Height) > Height then
    begin
      FScrolling := sdDown;
      Invalidate;
    end
    else if FSimpleControl.Top < 0 then
    begin
      FScrolling := sdUp;
      Invalidate;
    end
    else
      FScrolling := sdNotScrolling;
    // 15/10/99 DJ -------------------------------------------------------------
  end;
end;

{------------------------------------------------------------ Event Handlers --}

procedure TCustomTimePanel.DoTimeChange;
begin
  if Assigned(FOnTimeChange) then FOnTimeChange(Self);
end;

procedure Register;
begin
  RegisterComponents('Time', [TTimePanel]);
end;

end.
